#ifndef AutoSaveRestore_h_included_p
#define AutoSaveRestore_h_included_p

namespace cxxtls
{
     template<typename Value>
     struct AutoSaveRestore
         /// A helper object that lets you save and restore a memory location's 
         /// contents.  Construct an object of this type with a reference to a
         /// variable and during this object's destruction, it will be reset to
         /// its original value;
     {
         Value *ptr_;
         Value tmp_;

         AutoSaveRestore(Value &r)
         :  ptr_(&r)
         {
            tmp_ = *ptr_;
         }

         ~AutoSaveRestore()
         {
            *ptr_ = tmp_;
         }
     };
};

#endif
