#include <cxxtls/lastErrno.h>
#include <cxxtls/foreach.h>
#include <cxxtls/lexical_cast.h>

/////////////////////////////////////////////////////
#ifdef _MSC_VER
#include <lastWinErrStr.h>
#include <WinError.h>
#include <windows.h>
#include <map>

static std::map<int, std::string> errorTable;

static int populateErrorTable()
{
    CXXTLS_FOREACH(WinErrorNoTable const &cur, winErrorNoTable)
    {
       errorTable.insert( std::pair<int,std::string>(cur.id, cur.name) );
    }
    return 0;
}

static int populator = populateErrorTable();  // at static init time, populate variable errorTable

#else
#include <errno.h>
#include <string.h>
#endif
/////////////////////////////////////////////////////


std::string cxxtls::lastErrno() 
{
#ifdef _MSC_VER
    DWORD err = GetLastError();

    std::map<int, std::string>::const_iterator where = errorTable.find(err);

    if(where == errorTable.end())
    {
        std::string rv("Error:  ");

        rv += cxxtls::lexical_cast<std::string>(err);

        return rv;
    }

    return where->second;

#else
    int err = errno;

    char buffer[1000];

    snprintf(buffer, sizeof(buffer), "error:  %d", err);

    strerror_r(err, buffer, sizeof(buffer));

    buffer[sizeof(buffer)-1] = 0;

    return buffer;


#endif
}

