//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.
#include <cxxtls/scriptthread.h>
#include <cxxtls/scriptfunctions.h>
#include <cxxtls/foreach.h>
#include <cxxtls/thread.h>
#include <cxxtls/file.h>
#include <cxxtls/ScriptLanguageException.h>
#include <cxxtls/lastErrno.h>
#include <WinError.h>
#include <cxxtls/AutoDestroyingPointer.h>

#include <vector>

#ifndef _MSC_VER
#include <unistd.h>
#include <sys/stat.h>
#include <errno.h>
#endif

using namespace std;

namespace cxxtls
{
namespace ScriptThread
{

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
    ManagedFileHandles::
    ManagedFileHandles()
   {
       currentDirectory_ = FileName::getcwd();
   }
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
    ManagedFileHandles::
   ~ManagedFileHandles()
   {
       CXXTLS_FOREACH(HandlesVector::value_type &cur, handles_)
       {
           cur.second.closeIfNotError();
       }
   }

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

    ManagedFileHandles::FileHandleAttributes 
    ManagedFileHandles::getStdinFileHandleAttributes()
    {
       FileHandleAttributes rv;

       rv.fh_ = 0;
       rv.input_ = true;
       rv.error_ = false;

#ifdef _MSC_VER

       SetLastError(0);

       rv.msHandle_ = GetStdHandle(STD_INPUT_HANDLE);
       if (rv.msHandle_ == INVALID_HANDLE_VALUE)
       {
           rv.errno_ = GetLastError();
           rv.error_ = true;
           rv.msHandle_ = 0;
           return rv;
       }

       HANDLE tmp = 0;

       SetLastError(0);


       if(!DuplicateHandle(GetCurrentProcess(), 
                           rv.msHandle_, 
                           GetCurrentProcess(),
                           &tmp, 
                           0,
                           FALSE,
                           DUPLICATE_SAME_ACCESS
                          )
         )
       {
           rv.msHandle_ = 0;
           rv.error_ = true;
           rv.errno_ = GetLastError();
       }

       rv.msHandle_ = tmp;  // return the duplicated copy


#else
       rv.unixHandle_ = dup(0);

       if(rv.unixHandle_ < 0)
       {
          rv.errno_ = errno;
          rv.error_ = true;
          return rv;
       }
#endif


       return rv;
    }
                                                      

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
    ManagedFileHandles::FileHandleAttributes 
    ManagedFileHandles::getStdoutFileHandleAttributes()
    {
       FileHandleAttributes rv;

       rv.fh_ = 1;
       rv.input_ = false;
       rv.error_ = false;
       rv.errno_ = 0;

#ifdef _MSC_VER

       SetLastError(0);

       rv.msHandle_ = GetStdHandle(STD_OUTPUT_HANDLE);

       if (rv.msHandle_ == INVALID_HANDLE_VALUE)
       {
           rv.errno_ = GetLastError();
           rv.error_ = true;
           rv.msHandle_ = 0;
           return rv;
       }

       HANDLE tmp = 0;

       SetLastError(0);


       if(!DuplicateHandle(GetCurrentProcess(), 
                           rv.msHandle_, 
                           GetCurrentProcess(),
                           &tmp, 
                           0,
                           FALSE,
                           DUPLICATE_SAME_ACCESS
                          )
         )
       {
           rv.msHandle_ = 0;
           rv.error_ = true;
           rv.errno_ = GetLastError();
       }

       rv.msHandle_ = tmp;  // return the duplicated copy

#else
       rv.unixHandle_ = dup(1);

       if(rv.unixHandle_ < 0)
       {
           rv.errno_ = errno;
           rv.error_ = true;
           return rv;
       }
#endif


       return rv;

    }
                                                         

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
    ManagedFileHandles::FileHandleAttributes 
    ManagedFileHandles::getStderrFileHandleAttributes()
    {
       FileHandleAttributes rv;

       rv.fh_ = 2;
       rv.input_ = false;
       rv.error_ = false;

#ifdef _MSC_VER
       SetLastError(0);

       rv.msHandle_ = GetStdHandle(STD_ERROR_HANDLE);
       if (rv.msHandle_ == INVALID_HANDLE_VALUE)
       {
           rv.errno_ = GetLastError();
           rv.error_ = true;
           rv.msHandle_ = 0;
           return rv;
       }

       HANDLE tmp = 0;

       SetLastError(0);

       if(!DuplicateHandle(GetCurrentProcess(), 
                           rv.msHandle_, 
                           GetCurrentProcess(),
                           &tmp, 
                           0,
                           FALSE,
                           DUPLICATE_SAME_ACCESS
                          )
         )
       {
           rv.msHandle_ = 0;
           rv.error_ = true;
           rv.errno_ = GetLastError();
       }

       rv.msHandle_ = tmp;  // return the duplicated copy

#else
       rv.unixHandle_ = dup(2);

       if(rv.unixHandle_ < 0)
       {
          rv.errno_ = errno;
          rv.error_ = true;
          return rv;
       }
#endif

       return rv;
    }
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

    ManagedFileHandles::FileHandleAttributes 
    ManagedFileHandles::FileHandleAttributes::clone()
    {
       FileHandleAttributes rv = *this;
       rv.bytesInBuffer_ = 0;

       if(!error_)
       {
#ifdef _MSC_VER

           HANDLE tmp = 0;
           HANDLE org = rv.msHandle_;
           
           SetLastError(0);

           

           if(!DuplicateHandle(GetCurrentProcess(), 
                               rv.msHandle_, 
                               GetCurrentProcess(),
                               &tmp, 
                               0,
                               FALSE,
                               DUPLICATE_SAME_ACCESS
                              )
             )
           {
               rv.msHandle_ = 0;
               rv.error_ = true;
               rv.errno_ = GetLastError();

               if(rv.errno_)
               {
                   CXXTLS_LOG("clonehandle(" << org << ") failed");
               }


           }
           
           rv.msHandle_ = tmp;  // return the duplicated copy
#else
           rv.unixHandle_ = dup(rv.unixHandle_);

           if(rv.unixHandle_ < 0)
           {
               rv.msHandle_ = -1
               rv.error_ = true;
               rv.errno_ = errno;
           }
#endif
       }

       return rv;
    }

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
    ManagedFileHandles*  ManagedFileHandles::mainStdHandles = ManagedFileHandles::create();

#ifdef _MSC_VER

    // open function for nulHandle_

    static HANDLE nulHandleInitializer()
        ///  open a single, inheritable, nul file handle.  This file handle is basically an empty
        ///  source of input stream or sync of output stream.  This single file handle will not
        ///  be closed and can be inherited by called processes.
    {
        {
            HANDLE hFile;

            SECURITY_ATTRIBUTES  inheritable;
        
            inheritable.nLength = sizeof(inheritable);
            inheritable.bInheritHandle = true;           // harmless file to be always inherited  
            inheritable.lpSecurityDescriptor = 0;

            SetLastError(0);

            hFile = CreateFile("nul:",       // name of the file
                               GENERIC_ALL,            // open for writing
                               FILE_SHARE_READ|FILE_SHARE_WRITE,   // let others read/write
                               &inheritable,           // always inheritable.  CallProcess() clones needed files as inheritable.
                               OPEN_ALWAYS,
                               FILE_ATTRIBUTE_NORMAL,  // normal file
                               NULL);                  // no attr. template
        
            if (hFile == INVALID_HANDLE_VALUE) 
            { 
                // do not write an error message from a static initializer using CXXTLS_LOG!
                // 
                // returning invalid file handle seems to work ok

                // CXXTLS_LOG("CreateFile(nul:) error " << cxxtls::lastErrno() );
                // return 0;

            }

            return hFile;

        }

    }

    ManagedFileHandles::FileHandleAttributes::handle_t ManagedFileHandles::nulHandle_ = nulHandleInitializer();


#else
    ManagedFileHandles::FileHandleAttributes::handle_t nulHandle_ = open("/dev/null", O_RDWR);
#endif


///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
    static int standardHandleInitializer()
    {

       ManagedFileHandles::mainStdHandles->addFileHandle(ManagedFileHandles::getStdinFileHandleAttributes());
       ManagedFileHandles::mainStdHandles->addFileHandle(ManagedFileHandles::getStdoutFileHandleAttributes());
       ManagedFileHandles::mainStdHandles->addFileHandle(ManagedFileHandles::getStderrFileHandleAttributes());


       return 0;
    }

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
    static int callStandardHandleInitializer = standardHandleInitializer();
   

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
    void
    ManagedFileHandles::FileHandleAttributes::closeIfNotError()
    {
       if(fh_ >= 0)
       {
           // close the handle and set the error flag to true

#ifdef _MSC_VER
           
           if(isPipe_)
           {
              // Sleep(1);
           }

           if(pipeState_)
           {
               if (terminatePipe_) // true except when the while statement has call noTerminatePipe()
               {
               
                   pipeState_->setTerminated(true);  // signal to the other end of the pipe that the operations are terminated
                                                     // this is normally helpful only to the upstream member
               }
           }

           // CXXTLS_LOG(" closeHandle(" << fh_ << " [" << msHandle_ << "])");


#if 0
           if(!input_)
              FlushFileBuffers(msHandle_);   // bad!  when downstream pipes die first, this hangs! 
#endif

           SetLastError(0);

           CloseHandle(msHandle_);

           msHandle_ = 0;

           // the following just tells maintainers which side of the pipe set the termination flag.

           if(pipeState_)
           {
              if(fh_ == 1)
              {
                  pipeState_->setWriteHandle(0);
              }
              else
              {
                  if (terminatePipe_)
                      pipeState_->setReadHandle(0);   // do this unless we are in a while statement 
              }
           }

           DWORD err = GetLastError();

           if (err)
                CXXTLS_LOG("CloseHandle(" << fh_ << " [" << msHandle_ << "]" << ") Error:  " << cxxtls::lastErrno());

           msHandle_ = 0;
#else
           ::close(unixHandle_);
           unixHandle_ = -1;
#endif
           fh_ = -1;
       }

       error_ = true;
    }


   Mutex ManagedFileHandles::currentDirectoryMutex_;

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
    bool ManagedFileHandles::FileHandleAttributes::reopenAs(std::string const &filename_in, bool atEnd)
    {
        bytesInBuffer_ = 0;

        std::string filename = filename_in;

#ifdef _MSC_VER

        //
        //  windows implementation
        //
        //

        if (filename == "/dev/null")
            filename = "nul:";

        SECURITY_ATTRIBUTES  inheritable;

        inheritable.nLength = sizeof(inheritable);
        inheritable.bInheritHandle = false;             // Normal file handles, should NOT be inheritable.
        inheritable.lpSecurityDescriptor = 0;           // The callProcess function clones only the minimal
                                                        // subset of needed file handles and makes the copies
                                                        // inheritable.  
        
        if(input_)
        {
           HANDLE hFile;

           SetLastError(0);

           hFile = CreateFile(filename.c_str(),           // name of file to read
                                  GENERIC_READ,           // open for reading
                                  FILE_SHARE_READ,        // let others threads read too
                                  &inheritable,           // Do not pass this handle to child processes -- callProcess must clone as inheritable
                                  OPEN_EXISTING,          // 
                                  FILE_ATTRIBUTE_NORMAL,  // normal file
                                  NULL);                  // no attr. template
           
               if (hFile == INVALID_HANDLE_VALUE) 
               { 
                   CXXTLS_LOG("CreateFile(" << filename << ") error " << cxxtls::lastErrno() );
                   return false;
               }

               CloseHandle(msHandle_);
               msHandle_ = hFile;
        }
        else
        {
            HANDLE hFile;

            SetLastError(0);

            hFile = CreateFile(filename.c_str(),       // name of the file
                               GENERIC_WRITE,          // open for writing
                               FILE_SHARE_READ,        // let others read
                               &inheritable,           // always non-inheritable.  CallProcess() clones needed files as inheritable.
                               atEnd ? OPEN_ALWAYS : CREATE_ALWAYS, // open existing or non-existing files both
                               FILE_ATTRIBUTE_NORMAL,  // normal file
                               NULL);                  // no attr. template
        
            if (hFile == INVALID_HANDLE_VALUE) 
            { 
                CXXTLS_LOG("CreateFile(" << filename << ") error " << cxxtls::lastErrno() );
                return false;
            }

            CloseHandle(msHandle_);
            msHandle_ = hFile;

            error_ = false;

            if(atEnd)
            {
               // if we are supposed to seek to the end

               SetFilePointer(hFile,0,0,FILE_END);  // ignore errors on seek position.  if you can't seek, just don't.

            }
        }

#else
        //
        //  unix/linux implementation
        //
        //

        if(input_)
        {
            int fh = open(filename.c_str(), O_READ);

            if(fh < 0)
               return false;   // if error, keep the old handle open

            ::close(unixHandle_);

            unixHandle_ = fh;
        }
        else
        {
           int fh = open(filename.c_str(), O_CREAT|O_TRUNC, S_IREAD|S_IWRITE|S_EXEC);

           if(fh < 0)
              return false; // no can do big guy

           if(atEnd)
               lseek(fh, 0L, SEEK_END);

           ::close(unixHandle_);

           unixHandle_ = fh;

        }

#endif

        return true;
    }
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

   bool ManagedFileHandles::FileHandleAttributes::reopenDup(handle_t  os_handle)
   {
        bytesInBuffer_ = 0;

#ifdef _MSC_VER



      // CXXTLS_LOG("+reopenDup(" << os_handle << ")" );
      // CXXTLS_LOG("+  close(" << msHandle_ << ")"  );

       SetLastError(0);

      CloseHandle(msHandle_);

      if(GetLastError())
         CXXTLS_LOG("Error closing handle " << msHandle_);

      error_ = true;

      errno_ = -1;  // same value

       SetLastError(0);


      if(!DuplicateHandle(GetCurrentProcess(), 
                          os_handle, 
                           GetCurrentProcess(),
                           &msHandle_, 
                           0,
                           FALSE,
                           DUPLICATE_SAME_ACCESS
                          )
         )
       {

           msHandle_ = 0;
           error_ = true;
           errno_ = GetLastError();

           if(errno_ != 0)
                      CXXTLS_LOG("DuplicateHandle(" << fh_ << "[" << msHandle_ << "]" << ") Error:  " << cxxtls::lastErrno());

           CXXTLS_LOG("-reopenDup(" << os_handle << ") -- errout");

           return false;
       }
       else
       {
          error_ = false;

          errno_ = 0;
       }


       //CXXTLS_LOG("-reopenDup(" << os_handle << ") returns " << msHandle_ );

       return true;

#else
      ::close(unixHandle_);

      error_ = true;

      errno_ = -1;  // same value
      
      unixHandle_ = dup(os_handle);

      if(unixHandle_ < 0)
      {
          errno_ = errno;

          return false;
      }

      error_ = false;
      errno_ = 0;

      return true;


#endif
   }

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

    ManagedFileHandles::IO_ERROR
    ManagedFileHandles::writeBytes(int fh, char const *bytes, size_t length)
    {
        iterator where = find(fh);

        if(where == end())
            return IOERR_bad_file;

        if(where->second.input_)
           return IOERR_not_opened_for_write;

        if(where->second.pipeState_)
        {
            // the pipe creation tools, see BrashPipeStatement::execute(),
            // have created a  PipeStatus object for this handle.
            // 

            if(where->second.pipeState_->terminated())
                throw PipeWriteFailure();

            if(!where->second.pipeState_->hasReader(where->second.msHandle_))
            {
                // there is no downstream read handle for this upstream writer

                throw PipeWriteFailure();
            }


            // else, we are still reading/writing from the pipe as normal

        }


#ifdef _MSC_VER
        
        DWORD bytesWritten=0;

       SetLastError(0);

        if(!WriteFile(where->second.msHandle_,
                      bytes,
                      length,
                      &bytesWritten,
                      0
                     )

        )
        {
           DWORD err = GetLastError();

           CXXTLS_LOG("WriteFile(" << fh << ") error:  " << cxxtls::lastErrno() );

           if(   err == ERROR_OPERATION_ABORTED )
           {
               bool throwException =true;

               if(where->second.pipeState_)
               {
                  if(where->second.pipeState_->terminated())
                  {
                     err = ERROR_PIPE_NOT_CONNECTED;

                    throwException = false;
                  }
               }

               if(throwException)
                   throw UserIOCancellation();
           }

           if(
                 err == ERROR_BROKEN_PIPE
              || err == ERROR_PIPE_LOCAL
              || err == ERROR_BAD_PIPE
              || err == ERROR_PIPE_BUSY
              || err == ERROR_PIPE_NOT_CONNECTED
           )
               throw PipeWriteFailure(); // apparently the pipe close occurred during the write operation

           return IOERR_other_os_error;
        }

        if(bytesWritten != length)
        {
           CXXTLS_LOG( "Io Error on file handled " << fh << " Tried to write " << length << ", but only wrote " << bytesWritten << " bytes" );
        }


#else
        int err = write(fh, bytes, length);
#endif

        return IOERR_OK;

    }

static void   CallReadFile(HANDLE msHandle_,
                           char*  buffer_,
                           size_t bufsize,
                           DWORD  *inputBytes,
                           bool   *readResult,
                           bool   *SEH_ocurred
                          )
{
         *readResult = ReadFile(
                                             msHandle_,
                                             buffer_,
                                             bufsize,
                                             inputBytes,
                                             0
                              );


}

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

    ManagedFileHandles::IO_ERROR
    ManagedFileHandles::readBytes(int fh, size_t maxBytes, char *bytes, size_t *bytesRead)
    {
        iterator where = find(fh);

        if(where == end())
            return IOERR_bad_file;

        if(!where->second.input_)
           return IOERR_not_opened_for_read;

#ifdef _MSC_VER
        
        DWORD inputBytes=0;

        SetLastError(0);

        bool readResult = 0;
        bool SEH_occurred = false;

        CallReadFile(
                      where->second.msHandle_,
                      bytes,
                      maxBytes,
                      &inputBytes,
                      &readResult,
                      &SEH_occurred
                     );



        if(!readResult)
        {
           *bytesRead = inputBytes;

           DWORD err = GetLastError();

           if(err != ERROR_BROKEN_PIPE)
           {
               // not an error, just the way pipes work.
               CXXTLS_LOG("ReadFile(" << fh << ") failed: " << cxxtls::lastErrno() );
           }

           if(SEH_occurred)
              err = ERROR_OPERATION_ABORTED;

           if(err == ERROR_OPERATION_ABORTED)
               throw UserIOCancellation();


           return IOERR_other_os_error;
        }

        *bytesRead = inputBytes;

#else
        *bytesRead = read(fh, bytes, length);

        if(bytesRead < 0)
           return IOERR_other_os_error;
#endif

        return IOERR_OK;

    }


///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

    char const * ManagedFileHandles::errorString(ManagedFileHandles::IO_ERROR e)
    {
           switch(e)
           {
               case IOERR_OK                  : return "IOERR_OK";
               case IOERR_bad_file            : return "IOERR_bad_file";  
               case IOERR_closed_file         : return "IOERR_closed_file";
               case IOERR_not_opened_for_read : return "IOERR_not_opened_for_read";
               case IOERR_other_os_error      : return "IOERR_other_os_error";
               case IOERR_not_opened_for_write: return "IOERR_not_opened_for_write";
               
               default:                         return "<UNKNOWN>";
           }
    }

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
    bool ManagedFileHandles::reopenAs(int fh, std::string const &filename, bool atEnd)
    {

        iterator where = find(fh);

        if(where != end() && where->second.input_ && atEnd)
           return false;   // change append to a read file

        if(where == end())
        {
           // we are going to treat "reopen" as "open"

           FileHandleAttributes  tmp;

           tmp.fh_ = fh;
           tmp.error_ = true;
           tmp.errno_ = 0;
           tmp.input_ = false; // this needs to be a parameter

#ifdef _MSC_VER
           tmp.msHandle_ = 0;
#else
           tmp.unixHandle_ = -1;
#endif

           addFileHandle(tmp);

           where = find(fh);

           if (where == end())
               throw cxxtls::ScriptLanguage::Exception("ManagedFileHandles::reopenAs", "reason", "file handles list is corrupt");

        }

        {
            ManagedFileHandles::ScopedCurrentDirectoryLock lock;

            chdir(currentDirectory_.c_str());  // file may not be in currend directory...

            return where->second.reopenAs(filename, atEnd);
        }


    }


///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
    bool ManagedFileHandles::reopenDup(int modified_fh, int copy_from_fh)
    {
        

        iterator source = find(copy_from_fh);

        if(source == end())
            return false;

        handle_t handle = source->second.getOSHandle();

        iterator target = find(modified_fh);

        if(target != end())
            return target->second.reopenDup(handle);

        // creating a new one
        
        FileHandleAttributes newbie = source->second;
        
        newbie.fh_ = modified_fh;

#ifdef _MSC_VER

       SetLastError(0);

       if(!DuplicateHandle(GetCurrentProcess(), 
                           handle, 
                           GetCurrentProcess(),
                           &newbie.msHandle_, 
                           0,
                           FALSE,
                           DUPLICATE_SAME_ACCESS
                          )
         )
       {
           DWORD err = GetLastError();

           if(err)
               CXXTLS_LOG("DuplicateHandle(" << copy_from_fh << ") error:  " << cxxtls::lastErrno() );

           newbie.msHandle_ = 0;
           return false;
       }
        
#else

       newbie.unixHandle_ = dup(handle);

       if(newbie.unixHandle_ < 0)
          return false;
#endif

      addFileHandle(newbie);

      return true;  

    }


///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
    ManagedFileHandles *
    ManagedFileHandles::
    clone()
    {
        ManagedFileHandles *rv = new ManagedFileHandles;

        rv->currentDirectory_ = currentDirectory_;

        bool error = false;

        for(size_t i = 0; i < handles_.size(); ++i)
        {
           FileHandleAttributes p = handles_[i].clone();

           if(p.error_)
              error = true;

           rv->handles_.insert( HandlesVector::value_type(i, p) ); 
        }

        if(error)
        {
           delete rv;
           return 0;
        }

        return rv;
    }

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
    ManagedFileHandles *
    ManagedFileHandles::
    pipe()
    {
        ManagedFileHandles *rv = new ManagedFileHandles;

        bool error = false;

        for(size_t i = 0; i < handles_.size(); ++i)
        {
           FileHandleAttributes p = handles_[i].clone();

           if(p.error_)
              error = true;

           rv->handles_.insert( HandlesVector::value_type(i, p) ); 
        }

        if(error)
        {
           delete rv;
           return 0;
        }

        // at this point, we have duplicated clone().  The next step is to create the
        // pipe object and replace the standard output handle of the return value
        // with a clone of input handle to the pipe.  And in *this, replace the standard
        // input handle with the standard output handles from the pipe.

        SECURITY_ATTRIBUTES sa;

        sa.nLength               = sizeof(sa);
        sa.bInheritHandle        = false;            // launching child process will clone as inheritable as needed
        sa.lpSecurityDescriptor  = 0;

        HANDLE readHandle=0;
        HANDLE writeHandle=0;

       SetLastError(0);

        bool success = CreatePipe(
                                  &readHandle,
                                  &writeHandle,
                                  &sa,
                                  1 // use default buffer size
                                 );

        if(!success)
        {
           DWORD err = GetLastError();

           if(err)
               CXXTLS_LOG("CreatePipe() error:  " << cxxtls::lastErrno() );

           delete rv;
           return 0;
        }

        if(handles_.size() < 2)
        {
           DWORD err = GetLastError();
           CXXTLS_LOG("CreatePipe() error,too few handles produced:  " << cxxtls::lastErrno() );

           // should never happen
           delete rv;
           return 0;
        }


        CloseHandle(handles_[0].msHandle_);
        handles_[0].msHandle_ = readHandle;

        CloseHandle(rv->handles_[1].msHandle_);
        rv->handles_[1].msHandle_ = writeHandle;
        rv->handles_[1].isPipe_   = true;           


        return rv;
    }


///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

   static Mutex processCallGuard;

   int
   ManagedFileHandles::
   callProcess(std::string const &commandLine,
               StringExpr::ExecutionVariables const &env,
               bool isPipeMiddle
              )
   {
       std::map<int, HANDLE> clonedInheritableHandles;

       typedef std::map<int, HANDLE>::value_type MapMember;

       
       StringExpr::ExecutionVariables::ProcessVariableSet childEnvironmentPackage;
       
       env.getProcessVariables(childEnvironmentPackage);  // package variables for call

       int exitCode = -1;

#ifdef _MSC_VER
       PROCESS_INFORMATION processInfo;

#endif

       {

          // note:  the creation of the child process should be protected via a mutex, but not the waiting
          // for it to finish.

           ScopedMutexLock  lock(processCallGuard);                  // prevent other callers from tromping all over us while spawn our child
           ManagedFileHandles::ScopedCurrentDirectoryLock lock2;     // prevent others from changing the current directory

           chdir(currentDirectory_.c_str());


#ifdef _MSC_VER

           //   BOOL WINAPI CreateProcess(
           //     _In_opt_     LPCTSTR lpApplicationName,
           //     _Inout_opt_  LPTSTR lpCommandLine,
           //     _In_opt_     LPSECURITY_ATTRIBUTES lpProcessAttributes,
           //     _In_opt_     LPSECURITY_ATTRIBUTES lpThreadAttributes,
           //     _In_         BOOL bInheritHandles,
           //     _In_         DWORD dwCreationFlags,
           //     _In_opt_     LPVOID lpEnvironment,
           //     _In_opt_     LPCTSTR lpCurrentDirectory,
           //     _In_         LPSTARTUPINFO lpStartupInfo,
           //     _Out_        LPPROCESS_INFORMATION lpProcessInformation
           //   );

           size_t commandLineBufferSize = 1 + commandLine.size() ;
           
           char * commandLineBuffer = (char*)(operator new(commandLineBufferSize));
           
           memcpy(commandLineBuffer, commandLine.data(), commandLineBufferSize-1);
           
           commandLineBuffer[commandLineBufferSize-1] = 0;
       
       
           STARTUPINFO  startInfo;

           GetStartupInfo(&startInfo);

           //
           //  for security reasons, we do normally create files as inherited.  Instead,
           //  in that exact instant when we spawn a child process, we temporarily create
           //  inheritable clones of the handles that we wish to pass to the child process.
           //  then, as soon as we get the child created, we delete the clones so they won't
           //  accidentally get sent to the next child that gets created.
           // 
           //  Note that only 1 child process can be in this section of code at any one
           //  time because a scoped mutex lock prevents a second one from entering.
           //



           bool zeroFound = false;
           bool oneFound  = false;
           bool twoFound  = false;

           CXXTLS_FOREACH(HandlesVector::value_type & cur, handles_)
           {
               // clone the handles that are supposed to be inherited as inheritable.


               HANDLE tmp = 0;
        
               SetLastError(0);
        
        
               if(!DuplicateHandle(GetCurrentProcess(), 
                                   cur.second.getOSHandle(),
                                   GetCurrentProcess(),
                                   &tmp, 
                                   0,
                                   TRUE,
                                   DUPLICATE_SAME_ACCESS
                                  )
                 )
               {
                   DWORD err = GetLastError();

                   CXXTLS_LOG("Error " << err << ":  unable to clone handle " << cur.second.getOSHandle() << " for child process");

                   tmp = nulHandle_;

               }


               clonedInheritableHandles.insert(MapMember(cur.first, tmp) );

               if(cur.first == 0)
                  zeroFound = true;
               else
               if(cur.first == 1)
                  oneFound = true;
               else
               if(cur.first == 2)
                  twoFound = true;
           }




           if(zeroFound)
           {
               HANDLE tmp = clonedInheritableHandles.find(0)->second;
               startInfo.hStdInput  = tmp;
           }
           else
               startInfo.hStdInput  = nulHandle_;

           if(oneFound)
               startInfo.hStdOutput = clonedInheritableHandles.find(1)->second;
           else
               startInfo.hStdOutput = nulHandle_;
                
           if(twoFound)
               startInfo.hStdError  = clonedInheritableHandles.find(2)->second;
           else
               startInfo.hStdError  = nulHandle_;

           startInfo.dwFlags = STARTF_USESTDHANDLES;

   
           CXXTLS_DEBUG("callProcess(" << commandLine << ")" );

           bool processCreated =    CreateProcess(  0, // let CreateProcess parse the command line for us
                                                    commandLineBuffer,
                                                    0, // lpProcessAttributes
                                                    0, // lpThreadAttributes,
                                                    true, // bInheritHandles
                                                    0, // dwCreationFlags
                                                    childEnvironmentPackage.vars(),
                                                    0, // lpCurrentDirectory
                                                    &startInfo, // startupInfo
                                                    &processInfo
                                                 );

           
           operator delete(commandLineBuffer);



           if(!processCreated)
           {
              CXXTLS_DEBUG("failed to callProcess(" << commandLine << ")" );

              CXXTLS_FOREACH(MapMember &cur, clonedInheritableHandles)
                 CloseHandle(cur.second);


              return -1;
           }

           // create was successful - now close our copy of the inheritable files to prevent
           // other processes form accidentally inheriting the same files

          // now close all the file handles duplicated for use in calling the child process.
          
          CXXTLS_FOREACH(MapMember &cur, clonedInheritableHandles)
          {
              CloseHandle(cur.second);
          }


       }

      // Successfully created the process.  Wait for it to finish.
      WaitForSingleObject( processInfo.hProcess, INFINITE );

      CXXTLS_DEBUG("completed process(" << commandLine << ")" );

 
      // Get the exit code.

      DWORD myExitCode=-2;
      GetExitCodeProcess(processInfo.hProcess, &myExitCode);

      exitCode = myExitCode;
 
   
      
      // Close the handles.
      CloseHandle( processInfo.hProcess );
      CloseHandle( processInfo.hThread );
 




#else

    // on unix, do something like this:
    //
    //    dup2(handles_[0].unixHandle_, 0)
    //    dup2(handles_[1].unixHandle_, 1)
    //    dup2(handles_[2].unixHandle_, 2)
    //    spawn ....


#endif
         


       return exitCode;

   } // callProcess()

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

   ThreadContext *
   ThreadContext::
   cloneForSameJob(
         std::string const &scriptName,
         StringExpr::ExecutionContext::ArgsVector const &  argsVector
        ) 
   {
       AutoDestroyingPointer<ManagedFileHandles> clonedHandles(fileHandles_->clone());

       if (clonedHandles.get() == 0)
           return 0;

       ThreadContext* rv = 
       create(scriptName,
              argsVector,
              env_, 
              clonedHandles.get()
             );

       if (rv)
       {
           clonedHandles.release(); // now owned by rv
       }
       else
       {
           return 0;  // failed to allocate child context.  let teh clonedHandles destructor delete it for us
       }


       rv->variables().setLastParmOfLastCommand(variables().lastParmOfLastCommand());

       //         commandContext->variables().setLastParmOfLastCommand(myContext->variables().lastParmOfLastCommand());


       return rv;

   }

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

   ThreadContext *
   ThreadContext::
   pipe() 
   {
       AutoDestroyingPointer<ManagedFileHandles> newbieOwner( fileHandles_->pipe() );

       ThreadContext* rv = 0;

       ManagedFileHandles *newbie = newbieOwner.get();

       if(!newbie)
          throw UnableToCreatePipe();

       {
          // keep together for exception safety
          
           rv =  createForNewJob(variables().scriptName(),
                                                variables().arguments(),
                                                env_, 
                                                newbie  // if null, we won't get here
                                               );
           
           rv->variables().setGlobbing( variables().globbing() );

           newbieOwner.release();
       }


       cxxtls::ThreadSafeHandle<ScriptThread::PipeState> pipeState(  new PipeState );

       pipeState->setWriteHandle( newbie->getOSHandle(1) );
       pipeState->setReadHandle(  fileHandles_->getOSHandle(0) );

       newbie->setPipeState(1, pipeState);
       fileHandles_->setPipeState(0, pipeState);
                     
       return rv;

   }


///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

   ThreadContext *
   ThreadContext::
   cloneForNewJob() 
   {
       ThreadContext* rv =  createForNewJob(variables().scriptName(),
                                            variables().arguments(),
                                            env_, 
                                            fileHandles_->clone()
                                           );

       rv->variables().setGlobbing ( variables().globbing() );
                     
       return rv;

   }


///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

   CXXTLS_DECLARE_THREAD_FUNCTION(Job::start,jobPtr)
   {
       Job *job = (Job*)(jobPtr);

       long long rv = (long long)job->run();

       delete job;  // clean up after we leave.

       return rv;
   }


///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

        ThreadContext::
        ThreadContext(std::string const &                               scriptName,
                      StringExpr::ExecutionContext::ArgsVector const &  argsVector,
                      StringExpr::ExecutionVariables                   *startEnv, 
                      ManagedFileHandles                               *fileHandles /* ownership taken by this object*/ 
                     )
        :  clonedEnvironment_()
        ,  env_(startEnv)
        ,  varContext_(startEnv)
        ,  fileHandles_(fileHandles)
        ,  execLog_(false)
        {
           // construct a thread context using an existing string execution context

           // all these functions create atomic copies.  that is the copies have no pointers or references
           // outside themselves -- particularly not the std::Strings.

           // we need to copy the superseeds from the owning varContext into this one so that we
           // don't have change the actual environment just for those settings.

           varContext_.setArguments(argsVector);  
           varContext_.setScriptName(scriptName);
           varContext_.setThreadExecutionContext(this);
        }

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

        ThreadContext::
        ThreadContext(
                      StringExpr::ExecutionVariables                   &parentEnv, 
                      std::string const &                               scriptName,
                      StringExpr::ExecutionContext::ArgsVector const &  argsVector,
                      ManagedFileHandles                               *fileHandles /* ownership taken by this object*/ 
                     )
        :  clonedEnvironment_(new StringExpr::ExecutionVariables(parentEnv))
        ,  env_(clonedEnvironment_.get())
        ,  varContext_(clonedEnvironment_.get())
        ,  fileHandles_(fileHandles)
        ,  execLog_(false)
        {

           // construct a thread context using an newed-up string execution context

           // all these functions create atomic copies.  that is the copies have no pointers or references
           // outside themselves -- particularly not the std::Strings.

           // we need to copy the superseeds from the owning varContext into this one so that we
           // don't have change the actual environment just for those settings.

           varContext_.setArguments(argsVector);  
           varContext_.setScriptName(scriptName);
           varContext_.setThreadExecutionContext(this);
        }

    #ifdef _MSC_VER
    struct SaveRestoreConsoleMode
    {
        DWORD consoleMode_;

        HANDLE input_;


        SaveRestoreConsoleMode()
        {
            input_ = GetStdHandle(STD_INPUT_HANDLE);

            GetConsoleMode(input_, &consoleMode_);

            SetConsoleMode(input_, 503);   // basiscally normal input mode
        }

       ~SaveRestoreConsoleMode()
       {
          SetConsoleMode(input_, consoleMode_);
       }



    };
    #endif




    bool
    ManagedFileHandles::
    FileHandleAttributes::
    readLine(std::string &line)
    {
        #ifdef _MSC_VER

         auto_ptr<SaveRestoreConsoleMode> old_console_mode;

         
         {   
            DWORD tmp;

            if(GetConsoleMode(msHandle_, &tmp))
            {
               // this a console

               old_console_mode.reset( new SaveRestoreConsoleMode() );
            }

         }
        #endif


        line.resize(0);

        if(error_ || !input_)
           return false;

        if(bytesInBuffer_)
        {
           // there are remaining bytes in the buffer_, left over from the last read operation.

           char *where = (char*)memchr(buffer_, '\n', bytesInBuffer_);

           if(where)
           {
               // there is a newline in the data in the buffer, lets return that
               // and remove those bytes from the buffer.  Get rid of trailing \r as part of the process

               --where;

               bool incWhere = false;

               if (*where == '\r')
               {
               
                   *where = '\n';
                   incWhere = true;
               }
               else
                   ++where;   // \r was missing, ignore the situation -- especially on unix

               line.append(buffer_, where+1);   // use the data, include the \n in the data

               if (incWhere)
                   ++where;

               memcpy(buffer_, where+1, bytesInBuffer_ - (where-buffer_) - 1);  

               bytesInBuffer_ -= 1 + (where - buffer_);

               return true;

           }

           // if we get here, the buffer has data in it but not the end of the line.

           line.append(buffer_, &buffer_[bytesInBuffer_]);  // copy the data out of the buffer
                                                                  // into the target string.

           bytesInBuffer_ = 0;


        }

        // bytesInBuffer_ should be zero here or you broke the code...

        for(;;)
        {
            int bytesRead;

            if(bytesInBuffer_ != 0)
            {
                throw -1; // big time error
            }


#ifdef _MSC_VER
            
            DWORD inputBytes=0;
            
            bool readResult = 0;

            bool SEH_occurred = false;

            SetLastError(0);

            CallReadFile(
                                   msHandle_,
                                   buffer_,
                                   sizeof(buffer_),
                                   &inputBytes,
                                   &readResult,
                                   &SEH_occurred
                    );


            if(! readResult)
            {
               // io error means return.

           DWORD err = GetLastError();

           if(err != ERROR_BROKEN_PIPE)
           {
               CXXTLS_LOG("ReadFile failed in readLine(): " << cxxtls::lastErrno());
           }

           if(SEH_occurred)
              err = ERROR_OPERATION_ABORTED;

           if (err == ERROR_OPERATION_ABORTED)
               throw UserIOCancellation();

           if (terminatePipe_)
               error_ = true;   // normally we want to do this, but in while statements and for statements, we don't.  Multiple attempts to read the closed file should not crash the interpreter
                                // which it will if you do this.  Well, it'll throw an exception.

               if(line.empty())
                  return false;

               return true; // the end of file be handled on the next call
            }

            if(inputBytes == 0)
            {
                DWORD err = GetLastError();


                if (err == ERROR_OPERATION_ABORTED)
                    throw UserIOCancellation();

               error_ = true;
               return !line.empty();
            }

            bytesRead = inputBytes;
            
#else
            bytesRead = read(fh_, buffer_, sizeof(buffer_));

            if(bytesRead <= 0)
            {
               good_ = false;

               error_ = int(bytesRead) < 0;
               return !line.empty();
            }
#endif
            // if we get here, then is data in the buffer_.


            bytesInBuffer_ = bytesRead;  

           char *where = (char*)memchr(buffer_, '\n', bytesInBuffer_);

           if(where)
           {
               // there is a newline in the data in the buffer, lets return that
               // and remove those bytes from the buffer.

               bool incWhere = false;

               if(where > buffer_)
               {
                   --where;

                   if (*where == '\r')
                   {
                   
                       *where = '\n';
                       incWhere = true;
                   }
                   else
                      ++where;  // no \r before the \n
               }

               line.append(buffer_, where+1);   // use the data, include the \n in the data

               if (incWhere) 
                   ++where;

               memcpy(buffer_, where+1, bytesInBuffer_ - (where-buffer_) - 1);  // remove the line we are return from the buffer

               bytesInBuffer_ -= 1 + (where - buffer_);

               return true;

           }
           else
           {
              // data is in the buffer but there is no \n there -- so copy the data out and
              // reset the counters to 0

              line.append(buffer_, &buffer_[bytesInBuffer_]);

              bytesInBuffer_ = 0;

           }


        }



    }


   bool
   ManagedFileHandles::
   readLine(int fh, std::string &line)
   {
        iterator where = find(fh);

        if(where == end())
            return false;

        if(!where->second.input_)
           return false;

        return where->second.readLine(line);

   }

   ManagedFileHandles::FileHandleAttributes::handle_t
   ManagedFileHandles::getOSHandle(int fh)
   {
        iterator where = find(fh);

        if(where == end())
            return 0;


        return where->second.getOSHandle();

   }

   void
   ManagedFileHandles::setNoTerminatePipe(int fh, bool value)
   {
        iterator where = find(fh);

        if(where == end())
            return ;


        where->second.setNoTerminatePipe(value);

   }


   void
   ManagedFileHandles::setPipeState(int fh, cxxtls::ThreadSafeHandle<PipeState> &s)
   {
        iterator where = find(fh);

        if(where == end())
            return;


        where->second.setPipeState(s);

   }

   void
   ManagedFileHandles::terminatePipe(int fh)
   {
        iterator where = find(fh);

        if(where == end())
            return;

        if(where->second.pipeState_)
            where->second.pipeState_->setTerminated(true);

        // if this ain't a pipe, silently ignore it

   }


   bool
   ManagedFileHandles::
   FileHandleAttributes::
   isConsole() const
   {
      DWORD mode;

      if(GetConsoleMode(msHandle_, &mode))
         return true;

      return false;
   }




} // ScriptThread
} // cxxtls
